Imports System.Collections
Namespace System.Collections.Generic

   Public Class Enmrt
      ''' <param name="Root">Root-Knoten</param>
      ''' <param name="GetChilds">Delegat zur Ermittlung von Kind-Knoten</param>
      Public Shared Function FromRoot(Of T)( _
               ByVal Root As T, ByVal GetChilds As Func(Of T, IEnumerable)) As Enmrt(Of T)
         'einfacher Trick: Root als einziges Element einer Auflistung von T, nmlich eines T-Arrays
         Return New Enmrt(Of T)(New T() {Root}, GetChilds)
      End Function
   End Class

   ''' <summary>
   ''' erzeugt einen Enumerator, der hierarchische Datenstrukturen (Trees) enumeriert. 
   ''' Statt Rekursionen knnen so For-Each-Schleifen verwendet werden.
   ''' </summary>
   Public Class Enmrt(Of T) : Inherits Enumerable(Of T)

      Private _Roots As IEnumerable
      Private _GetChilds As Func(Of T, IEnumerable)

      'Da der Tree iterativ (statt rekursiv) durchgegangen wird, sind hier die Enumeratoren 
      'der Parent-Nodes zu speichern
      Private _ParentEnumerators As New Stack(Of IEnumerator)
      Private _BottomUp As Boolean
      Private _SkipChildren As Boolean
      Private _Movenext As Action
      Private _GetChildren As Func(Of IEnumerator)

      ''' <param name="Roots">Auflistung hierarchischer Objekte (Knoten)</param>
      ''' <param name="GetChilds">Delegat zur Ermittlung von Kind-Knoten</param>
      Public Sub New(ByVal Roots As IEnumerable, ByVal GetChilds As Func(Of T, IEnumerable))
         Me._Roots = Roots
         _GetChilds = GetChilds
         'Folgendes bringt einen interessanten Absturz der IDE:
         ''_GetChildren = Function() GetChilds(DirectCast(_Roots(0), T)).GetEnumerator
      End Sub

#Region "Publix"

      ''' <summary>
      ''' weist die Iteration an, die Kinder des aktuellen Knotens nicht zu durchlaufen
      ''' </summary>
      Public Sub SkipChildren()
         If _BottomUp Then Throw New Exception( _
               "Childnodes berspringen ist nur bei TopDown-Rekursion mglich.")
         _SkipChildren = True
      End Sub

      ''' <summary> Verschachtelungstiefe des aktuellen Knotens </summary>
      Public ReadOnly Property Depth() As Integer
         Get
            Return _ParentEnumerators.Count - 1
         End Get
      End Property

      ''' <summary>
      ''' emuliert eine "Bottom-Up"-Rekursion, d.h. die tiefst-verschachtelten Knoten werden 
      ''' zuerst behandelt
      ''' </summary>
      Public Property BottomUp() As Boolean
         Get
            Return _BottomUp
         End Get
         Set(ByVal value As Boolean)
            If _BottomUp = value Then Return
            If _ParentEnumerators.Count = 0 Then
               _BottomUp = value
            Else
               Throw New Exception( _
                  "Sie knnen die Rekursions-Reihenfolge nicht whrend der Rekursion ndern")
            End If
         End Set
      End Property

#End Region 'Publix

#Region "Overrides"

      Protected Overrides Sub Prepare()
         _Movenext = If(_BottomUp, _
               New Action(AddressOf MoveNextBottomUp), _
               New Action(AddressOf MoveNextTopDown))
         _GetChildren = If(_GetChilds Is Nothing, _
               Function() DirectCast(_ParentEnumerators.Peek.Current, IEnumerable).GetEnumerator, _
               Function() _GetChilds(DirectCast(_ParentEnumerators.Peek.Current, T)).GetEnumerator)
         _ParentEnumerators.Push(_Roots.GetEnumerator)
         _SkipChildren = True
         MyBase.Prepare()
      End Sub

      Protected Overrides Function TryMovenext(ByRef Current As T) As Boolean
         _Movenext()
         With _ParentEnumerators
            TryMovenext = .Count > 0
            If TryMovenext Then Current = DirectCast(.Peek.Current, T)
         End With
      End Function

      Protected Overrides Sub CleanUp()
         _ParentEnumerators.Clear()
         MyBase.CleanUp()
      End Sub

#End Region 'Overrides

#Region "Privates"

      Private Sub MoveNextTopDown()
         'Erst einen Child-Enumerator draufPushen.
         'Dann solange runter poppen, bis einer erfolgreich voranschreitet
         With _ParentEnumerators
            If _SkipChildren Then
               _SkipChildren = False
            Else
               .Push(_GetChildren()) ' (nach CleanUp oder Anforderung ggfs. auch berspringen)
            End If
            While .Count > 0 AndAlso Not .Peek.MoveNext
               .Pop()
            End While
         End With
      End Sub

      Private Sub MoveNextBottomUp()
         'Solange Child-Enumeratoren draufPushen und einen Schritt gehen lassen, bis das 
         'Voranschreiten fehlschlgt. Dann oberstes runternehmen.
         With _ParentEnumerators
            While .Peek.MoveNext
               .Push(_GetChildren())
            End While
            .Pop()
         End With
      End Sub

#End Region 'Privates

   End Class

End Namespace
